/*-----------------------------------------------------------------------------
 * $RCSfile: ShoppingClMain.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include "ShoppingClMsgMgr.h"

#include <OpenSOAP/OpenSOAP.h>
#include <OpenSOAP/StringHash.h>

#include <stdio.h>
#include <string.h>

#define APPLICATION_ERROR (0x10000001L)

static const unsigned char NULL_CHAR = '\0';

static const int FILE_DELM =
#if defined(__GNUC__)
'/';
#else
// 
'\\';
#endif

static const char DEFAULT_ENDPOINT[]
= "http://opensoap.jp/cgi-bin/soapInterface.cgi";

static const char DEFAULT_CHARENC[] = "UTF-8";

static const char OPTION_PREFIX[] = "-";

static const char OPTION_SOAPENDPOINT[] = "s";
static const char OPTION_SOAPACTION[] = "a";
static const char OPTION_CHARENC[] = "c";

static const char USAGE_FORMAT[] = "\
Usage: %s [-s endpoint] [-a soapaction] [-c char_enc] method [code] [qty]\n\
  -s endpoint    SOAP service endpoint URI\n\
  -a soapaction  value of SOAPAction HTTP-Header\n\
  -c char_enc    character encoding of request message (default: UTF-8)\n\
  method         GetProductList, GetStockQty, PlaceOrder[Async],\n\
                 or ConfirmOrder.\n\
  code           product code (for GetStockQty and PlaceOrder)\n\
  qty            quantity of products to order (for PlaceOrder)\n\
";

typedef  struct {
	char *appName;
	char *soapEndpoint;
	char *soapAction;
	char *charEnc;
	char *method;
	char *code;
	long qty;
} ClientVariables;

/* get name of this program */
static
char *
GetAppName(const char *argv0) {
	char *ret = NULL;
	size_t  appNameLen = 0;
	const char *appName = strrchr(argv0, FILE_DELM);
	
	if (appName) {
		++appName;
	}
	else {
		appName = argv0;
	}
	appNameLen = strlen(appName);
	ret = malloc(appNameLen + 1);
	
	return ret ? strcpy(ret, appName) : NULL;
}

/* initialize variable given as command-line arguments */
static
long
InitializeVariables(ClientVariables *appVars,
					int argc,
					char **argv) {
	long error = 0;
	const size_t OPTION_PREFIX_LEN = strlen(OPTION_PREFIX);
	char **avEnd = argv + argc;
	
	memset((void *)appVars, 0, sizeof(*appVars)); /* clear */
	
	appVars->appName = GetAppName(*argv);

	for (++argv; argv != avEnd; ++argv) {
		const char *optBody = *argv;
		if (strncmp(optBody, OPTION_PREFIX, OPTION_PREFIX_LEN) == 0) {
			size_t optBodyLen = 0;
			optBody += OPTION_PREFIX_LEN;
			optBodyLen = strlen(optBody);
			if (strncmp(OPTION_SOAPENDPOINT, optBody, optBodyLen) == 0
				&& !appVars->soapEndpoint) {
				++argv;
				if (argv == avEnd) {
					break;
				}
				appVars->soapEndpoint = strdup(*argv);
			}
			else if (strncmp(OPTION_SOAPACTION, optBody, optBodyLen) == 0
				&& !appVars->soapAction) {
				++argv;
				if (argv == avEnd) {
					break;
				}
				appVars->soapAction = strdup(*argv);
			}
			else if (strncmp(OPTION_CHARENC, optBody, optBodyLen) == 0
					 && !appVars->charEnc) {
				++argv;
				if (argv == avEnd) {
					break;
				}
				appVars->charEnc = strdup(*argv);
				if (!appVars->charEnc) {
					fprintf(stderr, "memory allocate error\n");
					error = APPLICATION_ERROR;
					argv = avEnd;
					break;
				}
			}
			else {
				argv = avEnd;
				break;
			}
		}
		else {
			break;
		}
	}

	if (argv != avEnd) {
		
		appVars->method = strdup(*argv);
		++argv;
		if (argv == avEnd) {
			if (strcmp(appVars->method, "GetProductList") != 0
				&& strcmp(appVars->method, "ConfirmOrder") != 0) {
				/* error for GetStockQty and PlaceOrder[Async] */
				fprintf(stderr, "arguments error: too few parameters\n");
				error = APPLICATION_ERROR;
			}
		}
		else {
			if (strcmp(appVars->method, "GetProductList") == 0){
				/* error for GetProductList */
				fprintf(stderr, "arguments error: too many parameters\n");
				error = APPLICATION_ERROR;
			}
			appVars->code = strdup(*argv);
			++argv;
			if (argv == avEnd) {
				if (strcmp(appVars->method, "GetStockQty") != 0) {
					/* error for PlaceOrder[Async] */
					fprintf(stderr, "arguments error: too few parameters\n");
					error = APPLICATION_ERROR;
				}
			}
			else {
				if (strcmp(appVars->method, "GetStockQty") == 0) {
					/* error for GetStockQty */
					fprintf(stderr, "arguments error: too many parameters\n");
					error = APPLICATION_ERROR;
				}
				appVars->qty = atol(*argv);
				++argv;
				if (argv != avEnd) {
					/* error for PlaceOrder */
					fprintf(stderr, "arguments error: too many parameters\n");
					error = APPLICATION_ERROR;
				}
			}
		}
	}
	else {
		fprintf(stderr, USAGE_FORMAT,
				appVars->appName ? appVars->appName : "CalcClient");
		exit(0);
	}
	
	if (!appVars->soapEndpoint) {
		appVars->soapEndpoint = strdup(DEFAULT_ENDPOINT);
	}
	
	if (!appVars->charEnc) {
		appVars->charEnc = strdup(DEFAULT_CHARENC);
		if (!appVars->charEnc) {
			fprintf(stderr, "memory allocate error\n");
			error = APPLICATION_ERROR;
		}
	}
	
	if (error) {
		fprintf(stderr, USAGE_FORMAT,
			   appVars->appName ? appVars->appName : "ShoppingClient");
	}
	
	return error;
}

/* main */
int
main(int argc,
	 char **argv) {
	ClientVariables appVars;
	
	long error = OpenSOAPInitialize(NULL);
	
	if (OPENSOAP_FAILED(error)) {
		fprintf(stderr,
				"OpenSOAP API Initialize Failed.\n"
				"OpenSOAP Error Code: 0x%04x\n",
				error);
		
		return 1;
	}

	error = InitializeVariables(&appVars, argc, argv);
	if (OPENSOAP_FAILED(error)) {
		fprintf(stderr,
				"Application Initialize Failed.\n"
				"OpenSOAP Error Code: 0x%04x\n",
				error);
		OpenSOAPUltimate();

		return 1;
	}
	
	fprintf(stderr,
			"SOAPEndpoint: %s\n"
			"SOAPAction: %s\n"
			"character encoding: %s\n"
			"method: %s\n"
			"code: %s\n"
			"qty: %ld\n",
			appVars.soapEndpoint ? appVars.soapEndpoint : "",
			appVars.soapAction ? appVars.soapAction : "",
			appVars.charEnc ? appVars.charEnc : "",
			appVars.method,
			appVars.code ? appVars.code : "" ,
			appVars.qty);
	
	if (strcmp(appVars.method, "GetProductList") == 0) {
		ProductListPtr productList;
		productList = GetProductList(appVars.soapEndpoint,
									  appVars.soapAction,
									  appVars.charEnc);
		if (productList && productList->productCount) {
			ProductListItemPtr products = productList->productList;
			size_t i = 0;
			fprintf(stderr,
					"GetProductListResponse:\n"
					"(Manufacturer), (Name), (Code), (Price)\n");
			for (; i < productList->productCount; ++i) {
				fprintf(stderr,
						"%s, %s, %s, %ld\n",
						products[i].manufacturer,
						products[i].name,
						products[i].code,
						products[i].price);
			}
			fprintf(stderr, "\n");
			ReleaseProductList(productList);
		}
	}
	else if (strcmp(appVars.method, "GetStockQty") == 0) {
		long qty = -1;
		int error = GetStockQty(appVars.soapEndpoint,
								appVars.soapAction,
								appVars.charEnc,
								appVars.code,
								&qty);
		fprintf(stderr, "\nGetStockQtyResponse: qty=%ld\n\n", qty);
	}
	else if (strcmp(appVars.method, "PlaceOrder") == 0) {
		char* confirmation = NULL;
		int error = PlaceOrder(appVars.soapEndpoint,
							   appVars.soapAction,
							   appVars.charEnc,
							   appVars.code,
							   appVars.qty,
							   &confirmation);
		fprintf(stderr, "\nPlaceOrderResponse: %s\n\n", confirmation);
		free(confirmation);
	}
	else if (strcmp(appVars.method, "PlaceOrderAsync") == 0) {
		char* confirmation = NULL;
		int error = PlaceOrderAsync(appVars.soapEndpoint,
									appVars.soapAction,
									appVars.charEnc,
									appVars.code,
									appVars.qty,
									stdout);
	}
	else if (strcmp(appVars.method, "ConfirmOrder") == 0) {
		char* confirmation = NULL;
		int error = ConfirmOrder(stdin,
								 appVars.soapEndpoint,
								 appVars.soapAction,
								 appVars.charEnc,
								 &confirmation);
		fprintf(stderr, "\nConfirmOrderResponse: %s\n\n", confirmation);
	}

	OpenSOAPUltimate();
	
	return 0;
}
