/*-----------------------------------------------------------------------------
 * $RCSfile: ShoppingService.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>

#include "OpenSOAP/OpenSOAP.h"
#include "OpenSOAP/Service.h"

static const char SHOPPING_LOG_FILE[] = "/var/tmp/OpenSOAP/log/Shopping.log";

#if defined(CONNECT_TYPE_CGI)
# define CONNECT_TYPE "cgi"
# define LOG_STREAM stderr
#else
# define CONNECT_TYPE "stdio"
FILE *fp;
# define LOG_STREAM fp
#endif /* CONNECT_TYPE_CGI */

/* error message macro */
#define ERROR_MSG(error, location, message); \
fprintf(LOG_STREAM, \
		"%s: Cannot %s\n" \
		" ---> OpenSOAP Error Code: 0x%04x\n", \
		(location),\
		(message),\
		(error));

/* error check macro */
#define ERROR_CHECK(error, location, message); \
if (OPENSOAP_FAILED((error))) { \
	ERROR_MSG(error, location, message) \
}

/* error return macro */
#define ERROR_RETURN(error, location, message); \
if (OPENSOAP_FAILED((error))) { \
	ERROR_MSG(error, location, message) \
	return (error); \
}

static const char METHOD_NAMESPACE_URI[]
="http://services.opensoap.jp/samples/Shopping/";

static const char METHOD_NAMESPACE_PREFIX[] = "m";

typedef 
int
(*FunctionType)(OpenSOAPEnvelopePtr request,
				OpenSOAPEnvelopePtr *response);

typedef struct {
    const char *requestName;
    const char *responseName;
    FunctionType methodFunc;
} MethodMapItem;

typedef struct {
	const char *manufacturer;
	const char *name;
	const char *code;
	const long price;
} ProductSpec;

static
const
ProductSpec
ProductSpecList[] = {
	{ "NISSAN", "SKYLINE",  "NS01", 265 }, /* 00 */
	{ "TOYOTA", "VITZ",     "TY01",  84 }, /* 01 */
	{ "HONDA",  "STEPWGN",  "HN01", 185 }, /* 02 */
	{ "SUBARU", "LEGACY",   "SB01", 190 }, /* 03 */
	{ "HONDA",  "STREAM",   "HN02", 158 }, /* 04 */
	{ "SUZUKI", "WGN-R",    "SZ01",  87 }, /* 05 */
	{ "TOYOTA", "FUNCARGO", "TY02", 124 }, /* 06 */
	{ "HONDA",  "ODDESSAY", "HN03", 212 }, /* 07 */
	{ "NISSAN", "MARCH",    "NS02",  90 }, /* 08 */
	{ "SUBARU", "IMPREZZA", "SB02", 126 }, /* 09 */
	{ "NISSAN", "PREMERA",  "NS03", 204 }, /* 10 */
	{ "NISSAN", "SERENA",   "NS04", 189 }, /* 11 */
	{ "MAZDA",  "DEMIO",    "MZ01",  99 }, /* 12 */
	{ "TOYOTA", "IPSAM",    "TY03", 204 }, /* 13 */
	{ "TOYOTA", "BB",       "TY04", 129 }, /* 14 */
	{ "TOYOTA", "ESTIMA",   "TY05", 229 }, /* 15 */
	{ "MAZDA",  "MPV",      "MZ02", 207 }, /* 16 */
};

typedef struct {
	const char *code;
	const long qty;
} ProductStockQty;

static
const
ProductStockQty
ProductStockQtyList[] = {
	{ "NS01", 6 }, /* 00 */
	{ "TY01", 8 }, /* 01 */
	{ "HN01", 8 }, /* 02 */
	{ "SB01", 9 }, /* 03 */
	{ "HN02", 5 }, /* 04 */
	{ "SZ01", 8 }, /* 05 */
	{ "TY02", 2 }, /* 06 */
	{ "HN03", 1 }, /* 07 */
	{ "NS02", 9 }, /* 08 */
	{ "SB02", 2 }, /* 09 */
	{ "NS03", 0 }, /* 10 */
	{ "NS04", 8 }, /* 11 */
	{ "MZ01", 9 }, /* 12 */
	{ "TY03", 0 }, /* 13 */
	{ "TY04", 2 }, /* 14 */
	{ "TY05", 2 }, /* 15 */
	{ "MZ02", 0 }, /* 16 */
};

/* get parameter value of string type from message */
static
int
GetStringParameter(OpenSOAPBlockPtr body,
				   const char *paramName,
				   char **paramValue) {
	static char FuncName[] = "GetStringParameter";
	int error = OPENSOAP_NO_ERROR;
	OpenSOAPStringPtr param = NULL;
	size_t paramLen = 0;
	
	error = OpenSOAPBlockGetChildValueMB(body, paramName, "string", &param);
	ERROR_RETURN(error, FuncName, "get value of parameter");
	
	error = OpenSOAPStringGetLengthMB(param, &paramLen);
	ERROR_RETURN(error, FuncName, "get string length of parameter");
	
	*paramValue = malloc(paramLen + 1);	
	error = (*paramValue ? error : OPENSOAP_MEM_BADALLOC);
	ERROR_RETURN(error, FuncName, "allocate memory");
	
	error = OpenSOAPStringGetStringMB(param, &paramLen, *paramValue);
	ERROR_RETURN(error, FuncName, "get string from parameter");
	
	return error;
}

/* parse common part of request message */
static
int
ParseRequestCommon(OpenSOAPEnvelopePtr request,
				   const char *requestName,
				   OpenSOAPBlockPtr *body) {
	static char FuncName[] = "ParseRequestCommon";
	int error = OPENSOAP_NO_ERROR;
	int isSameNs = 0;
	
	error = OpenSOAPEnvelopeGetBodyBlockMB(request, requestName, body);
	ERROR_RETURN(error, FuncName, "get body block");
	
	error = OpenSOAPBlockIsSameNamespaceMB(*body,
										   METHOD_NAMESPACE_URI,
										   &isSameNs);
	ERROR_RETURN(error, FuncName, "complete namespace matching");
	
	error = (isSameNs ? error : OPENSOAP_NOT_CATEGORIZE_ERROR);
	ERROR_RETURN(error, FuncName, "match namespace");
	
	return error;
}

/* create common part of response message */
static
int
CreateResponseCommon(OpenSOAPEnvelopePtr *response,
					 const char *responseName,
					 OpenSOAPBlockPtr *body) {
	static char FuncName[] = "CreateRequestCommon";
	int error = OPENSOAP_NO_ERROR;
	
	error = OpenSOAPEnvelopeCreateMB("1.1", NULL, response);
	ERROR_RETURN(error, FuncName, "create envelope");
	
	/* add body response block */
	error = OpenSOAPEnvelopeAddBodyBlockMB(*response, responseName, body);
	ERROR_RETURN(error, FuncName, "add body part");
	
	/* set namespace to response block */
	error = OpenSOAPBlockSetNamespaceMB(*body,
										METHOD_NAMESPACE_URI,
										METHOD_NAMESPACE_PREFIX);
	ERROR_RETURN(error, FuncName, "set namespace of body part");
	
	return error;
}

/* create fault response message */
static
int
CreateFaultMessage(OpenSOAPEnvelopePtr *response,
				   const char *faultcode,
				   const char *faultstring,
				   const char *faultactor,
				   const char *detail) {
	static char FuncName[] = "CreateFaultMessage";
	int error = OPENSOAP_NO_ERROR;
	OpenSOAPBlockPtr body = NULL;
	OpenSOAPStringPtr faultcodeOssp = NULL;
	OpenSOAPStringPtr faultstringOssp = NULL;
	OpenSOAPStringPtr faultactorOssp = NULL;
	OpenSOAPStringPtr detailOssp = NULL;
	
	error = CreateResponseCommon(response, "Fault", &body);
	ERROR_RETURN(error, FuncName, "create common part of response");
	
	error = OpenSOAPStringCreateWithMB(faultcode, &faultcodeOssp);
	ERROR_RETURN(error, FuncName, "create string: faultcode");
	
	error = OpenSOAPBlockSetChildValueMB(body,
										 "faultcode",
										 "string",
										 &faultcodeOssp);
	ERROR_RETURN(error, FuncName, "set parameter: faultcode");
	
	error = OpenSOAPStringCreateWithMB(faultstring, &faultstringOssp);
	ERROR_RETURN(error, FuncName, "create string: faultstring");
	
	error = OpenSOAPBlockSetChildValueMB(body,
										 "faultstring",
										 "string",
										 &faultstringOssp);
	ERROR_RETURN(error, FuncName, "set parameter: faultstring");
	
	if (faultactor != NULL) {
		error = OpenSOAPStringCreateWithMB("faultactor", &faultactorOssp);
		ERROR_RETURN(error, FuncName, "create string: faultactor");
		
		error = OpenSOAPBlockSetChildValueMB(body,
											 "faultactor",
											 "string",
											 &faultactorOssp);
		ERROR_RETURN(error, FuncName, "set parameter: faultactor");
	}
	   
	if (detail != NULL) {
		
		error = OpenSOAPStringCreateWithMB(detail, &detailOssp);
		ERROR_RETURN(error, FuncName, "create string: detail");
		
		error = OpenSOAPBlockSetChildValueMB(body,
											 "detail",
											 "string",
											 &detailOssp);
		ERROR_RETURN(error, FuncName, "set parameter: detail");
	}
	
	return error;
}

/* process get product count request */
static
int
GetProductCount(OpenSOAPEnvelopePtr request,
				OpenSOAPEnvelopePtr *response) {
	static char FuncName[] = "GetProductCount";
	int error = OPENSOAP_NO_ERROR;
	long count = -1;
	OpenSOAPBlockPtr body = NULL;
	
	/* --- parse request message --- */
	
	error = ParseRequestCommon(request, "GetProductCount", &body);
	if (OPENSOAP_FAILED(error)) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "cannont parse common part of "
								   "GetProductCount request",
								   NULL,
								   NULL);
		return error;
	}
   
	/* --- make response contents --- */
	
	count = (long)(sizeof(ProductSpecList) / sizeof(ProductSpec));
	
	/* --- create request message --- */
	
	error = CreateResponseCommon(response, "GetProductCountResponse", &body);
	ERROR_RETURN(error, FuncName, "create common part of response");
	
	error = OpenSOAPBlockSetChildValueMB(body, "count", "int", &count);
	ERROR_RETURN(error, FuncName, "set parameter: count");
	
	return error;
}

/* process get product specification request */
static
int
GetProductSpec(OpenSOAPEnvelopePtr request,
			   OpenSOAPEnvelopePtr *response) {
	
	static char FuncName[] = "GetProductSpec";
	int error = OPENSOAP_NO_ERROR;
	
	long index = -1;
	OpenSOAPStringPtr manufacturer = NULL;
	OpenSOAPStringPtr name = NULL;
	OpenSOAPStringPtr code = NULL;
	long price = -1;
	
	long count = (long)(sizeof(ProductSpecList) / sizeof(ProductSpec));
	
	OpenSOAPBlockPtr body = NULL;
	
	/* --- parse request message --- */
	
	error = ParseRequestCommon(request, "GetProductSpec", &body);
	if (OPENSOAP_FAILED(error)) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "cannont parse common part of "
								   "GetProductSpec request",
								   NULL,
								   NULL);
		return error;
	}
	
  	error = OpenSOAPBlockGetChildValueMB(body, "index", "int", &index);
	if (OPENSOAP_FAILED(error)) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "cannont get index parameter",
								   NULL,
								   NULL);
		return error;
	}
	
	/* --- make response contents --- */
	
	if (index < 0 || index >= count) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "index parameter out of range",
								   NULL,
								   NULL);
		return error;
	}
	
	error = OpenSOAPStringCreateWithMB(ProductSpecList[index].manufacturer,
									   &manufacturer);
	ERROR_RETURN(error, FuncName, "create string: manufacturer");
	
	error = OpenSOAPStringCreateWithMB(ProductSpecList[index].name, &name);
	ERROR_RETURN(error, FuncName, "create string: name");
	
	error = OpenSOAPStringCreateWithMB(ProductSpecList[index].code, &code);
	ERROR_RETURN(error, FuncName, "create string: code");
	
	price = ProductSpecList[index].price;
	
	/* --- create response message --- */
	
	error = CreateResponseCommon(response, "GetProductSpecResponse", &body);
	ERROR_RETURN(error, FuncName, "create common part of response");
	
	error = OpenSOAPBlockSetChildValueMB(body,
										 "manufacturer",
										 "string",	
										 &manufacturer);
	ERROR_RETURN(error, FuncName, "set parameter: manufacturer");
	
	error = OpenSOAPBlockSetChildValueMB(body, "name", "string", &name);
	ERROR_RETURN(error, FuncName, "set parameter: name");
	
	error = OpenSOAPBlockSetChildValueMB(body, "code", "string", &code);
	ERROR_RETURN(error, FuncName, "set parameter: code");
	
	error = OpenSOAPBlockSetChildValueMB(body, "price", "int", &price);
	ERROR_RETURN(error, FuncName, "set parameter: price");
	
	return error;
}

/* process get stock quantity request */
static
int
GetStockQty(OpenSOAPEnvelopePtr request,
			OpenSOAPEnvelopePtr *response) {
	static char FuncName[] = "GetStockQty";
	int error = OPENSOAP_NO_ERROR;
	
	char *code = NULL;
	long qty = -1;
	
	const int stockProductCount
		= (int)(sizeof(ProductStockQtyList) / sizeof(ProductStockQty));
	
 	int i = 0;
	
	OpenSOAPBlockPtr body = NULL;
	
	/* --- parse request message --- */
	
	error = ParseRequestCommon(request, "GetStockQty", &body);
	if (OPENSOAP_FAILED(error)) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "cannont parse common part of "
								   "GetStockQty request",
								   NULL,
								   NULL);
		return error;
	}
	
	error = GetStringParameter(body, "code", &code);
	if (OPENSOAP_FAILED(error)) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "cannont get code parameter",
								   NULL,
								   NULL);
		return error;
	}
	
	/* --- make response contents --- */
	
	while (i < stockProductCount
		   && strcmp(code, ProductStockQtyList[i].code) != 0) {
		i++;
	}
	if (i == stockProductCount) {
		qty = (long)(-1);
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "invalid code parameter",
								   NULL,
								   NULL);
		return error;
	}
	
	qty = (long)(ProductStockQtyList[i].qty);
	
	free(code);
	
	/* --- create response message --- */
	
	error = CreateResponseCommon(response, "GetStockQtyResponse", &body);
	ERROR_RETURN(error, FuncName, "create common part of response");
	
	error = OpenSOAPBlockSetChildValueMB(body, "qty", "int", &qty);
	ERROR_RETURN(error, FuncName, "set parameter: qty");
	
	return error;
}

/* process place order request */
static
int
PlaceOrder(OpenSOAPEnvelopePtr request,
		   OpenSOAPEnvelopePtr *response) {
	static char FuncName[] = "PlaceOrder";
	int error = OPENSOAP_NO_ERROR;
	
	char *code = NULL;
	long qty = -1;
	OpenSOAPStringPtr confirmation = NULL;
	OpenSOAPBlockPtr body = NULL;
	
	const int stockProductCount
		= (int)(sizeof(ProductStockQtyList) / sizeof(ProductStockQty));

	int i = 0;
	
	/* --- parse request message --- */
	
	error = ParseRequestCommon(request, "PlaceOrder", &body);
	if (OPENSOAP_FAILED(error)) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "cannont parse common part of "
								   "PlaceOrder request",
								   NULL,
								   NULL);
		return error;
	}
	
	error = GetStringParameter(body, "code", &code);
	if (OPENSOAP_FAILED(error)) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "cannont get code parameter",
								   NULL,
								   NULL);
		return error;
	}
	
	error = OpenSOAPBlockGetChildValueMB(body, "qty", "int", &qty);
	if (OPENSOAP_FAILED(error)) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "cannont get qty parameter",
								   NULL,
								   NULL);
		return error;
	}
	
	/* --- make response contents --- */
	
	while (i < stockProductCount
		   && strcmp(code, ProductStockQtyList[i].code) != 0) {
		i++;
	}
	
	if (i == stockProductCount) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "invalid code parameter",
								   NULL,
								   NULL);
		return error;
	}
	if (qty < 1 || qty > ProductStockQtyList[i].qty) {
		error = CreateFaultMessage(response,
								   "SOAP-ENV:service",
								   "qty parameter out of range",
								   NULL,
								   NULL);
		return error;
	}
	
	error = OpenSOAPStringCreateWithMB("order successfully placed",
									   &confirmation);
	ERROR_RETURN(error, FuncName, "create string: confirmation");
	
	free(code);
	
	/* --- create response message --- */
	
	error = CreateResponseCommon(response, "PlaceOrderResponse", &body);
	ERROR_RETURN(error, FuncName, "create common part of response");
	
	error = OpenSOAPBlockSetChildValueMB(body,
										 "confirmation",
										 "string",
										 &confirmation);
	ERROR_RETURN(error, FuncName, "set parameter: confirmation");
	
	return error;
}

static
const MethodMapItem
MethodMap[] = {
	{ "GetProductCount", "GetProductCountResponse", GetProductCount },
	{ "GetProductSpec",  "GetProductSpecResponse",  GetProductSpec  },
	{ "GetStockQty",     "GetStockQtyResponse",     GetStockQty     },
	{ "PlaceOrder",      "PlaceOrderResponse",      PlaceOrder      },
	{ NULL,              NULL,                      NULL            }
};

/* service function to be registered */
static
int
ServiceFunc(OpenSOAPEnvelopePtr request,
			OpenSOAPEnvelopePtr *response,
			void *opt) {
	static char FuncName[] = "ServiceFunc";
	int error = OPENSOAP_NO_ERROR;
	
	MethodMapItem *methodMapItem = (MethodMapItem *)opt;
	
	if (methodMapItem) {
		error = (methodMapItem->methodFunc)(request, response);
		ERROR_RETURN(error, FuncName, "complete method");
		/*
		fprintf(stderr, "method name: %s\n", methodMapItem->requestName);
		*/
	}
	
	return error;
}

/* main */
main(void) {
	static char FuncName[] = "main";
	int error = OPENSOAP_NO_ERROR;
	OpenSOAPServicePtr service = NULL;
	
#if !defined(CONNECT_TYPE_CGI)
	fp = fopen(SHOPPING_LOG_FILE, "a");
	fprintf(fp, "start logging...\n");
#endif /* CONNECT_TYPE_CGI */
	
	/* initialize service */
	error = OpenSOAPInitialize(NULL);
	ERROR_RETURN(error, FuncName, "initialize service");;
	
	/* create service */
	error = OpenSOAPServiceCreateMB(&service,
									"ShoppingService",
									CONNECT_TYPE,
									0);
	if (OPENSOAP_SUCCEEDED(error)) {
		const MethodMapItem *methodMapItem = MethodMap;
		
		/* register methods to service */
		for (; OPENSOAP_SUCCEEDED(error) && methodMapItem->requestName;
			 ++methodMapItem) {
			error = OpenSOAPServiceRegisterMB(service,
											  methodMapItem->requestName,
											  ServiceFunc,
											  (void *)methodMapItem);
			ERROR_CHECK(error, FuncName, "register method");
			/*
			fprintf(stderr, "method name: %s\n", methodMapItem->requestName);
			*/
		}
		
		if (OPENSOAP_SUCCEEDED(error)) {
			/* activate service */
			error = OpenSOAPServiceRun(service);
			ERROR_CHECK(error, FuncName, "run service");
		}
		
		/* release service */
		error = OpenSOAPServiceRelease(service);
		ERROR_CHECK(error, FuncName, "release");
	}
	else {
		ERROR_MSG(error, FuncName, "create");
	}
	
	/* finalize service */
	OpenSOAPUltimate();

#if !defined(CONNECT_TYPE_CGI)
	fclose(fp);
#endif /* CONNECT_TYPE_CGI */
	
	return error;
}
