/*-----------------------------------------------------------------------------
 * $RCSfile: sec_rsasig.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */
#include <OpenSOAP/Security.h>
#include "security_defs.h"
#include <openssl/rsa.h>
#include <openssl/pem.h>
#if defined(DEBUG)
#include <openssl/err.h>
#endif

/***************************************************************************** 
    Function      : Convert Hash Type (OPENSOAP_HA_* --> NID_*)
    Return        : 
 ************************************************ Yuji Yamawaki 01.09.10 *****/
static int convType(int iTypeIn)
{
    switch (iTypeIn) {
    case OPENSOAP_HA_MD5:
        return NID_md5;
    case OPENSOAP_HA_RIPEMD:
        return NID_ripemd160;
    default:                    /* OPENSOAP_HA_SHA */
        return NID_sha1;
    }
}
/***************************************************************************** 
    Function      : Generate Digital Signature(RSA)
    Return        : char*(Signature<base64>, NULL on Error Occur)
 ************************************************ Yuji Yamawaki 01.07.26 *****/
int
openSOAPSecMakeRSASign
(int                  iType,     /* (i)  Hash Type(OPENSOAP_HA_*) */
 const unsigned char* szData,    /* (i)  Data to be signed */
 unsigned long        ulDataLen, /* (i)  Length of Data */
 FILE*                fpPrivKey, /* (i)  RSA Private Key File stream */
 char**               pszDSig,   /* (o)  Digital Signature<base64> */
 char**               pszDigest) /* (o)  Digest Value<base64> */
                                 /*      (NULL is allowed) */
{
    RSA*           pRsa = NULL;
    char*          szSig = NULL;
    char*          szResBase64 = NULL;
    unsigned char* szHash = NULL;
    unsigned long  ulLenHash;
    int            nSigLen;
    int            nRet = OPENSOAP_NO_ERROR;
    fpos_t         fposKey;
    /* Check Arguments */
    if (szData == NULL || fpPrivKey == NULL || ulDataLen <= 0 ||
        pszDSig == NULL) {
        return OPENSOAP_PARAMETER_BADVALUE;
    }
    /* Backup Stream's Position */
    if (fgetpos(fpPrivKey, &fposKey) != 0) {
        nRet = OPENSOAP_IO_READ_ERROR;
        goto FuncEnd;
    }
    /* Hash Original Data */
    nRet = openSOAPSecMakeHash(iType, szData, ulDataLen,
                               &ulLenHash, &szHash);
    if (OPENSOAP_FAILED(nRet)) {
        return nRet;
    }
    /* Convert Hash(Digest) Value to base64 string */
    if (pszDigest != NULL) {
        nRet = openSOAPSecEncByBase64(szHash, ulLenHash, pszDigest);
        if (OPENSOAP_FAILED(nRet)) {
            goto FuncEnd;
        }
    }
    /* Generate RSA from Private key */
    if (PEM_read_RSAPrivateKey(fpPrivKey, &pRsa, NULL, NULL) == NULL) {
        nRet = OPENSOAP_SEC_SIGNGEN_ERROR;
        goto FuncEnd;
    }
    /* Recover File Position */
    if (fsetpos(fpPrivKey, &fposKey) != 0) {
        nRet = OPENSOAP_IO_READ_ERROR;
        goto FuncEnd;
    }
    /* Allocate Signature Area */
    if ((szSig = (char*)malloc(RSA_size(pRsa))) == NULL) {
        nRet = OPENSOAP_MEM_BADALLOC;
        goto FuncEnd;
    }
    /* Create Signature */
    if (RSA_sign(convType(iType),
                 szHash,
                 ulLenHash,
                 szSig,
                 &nSigLen,
                 pRsa) != 1) {
#if defined(DEBUG)
        char          szBuf[256];
        ERR_error_string(ERR_get_error(), szBuf);
        fprintf(stderr, "%s\n", szBuf);
#endif
        free(szSig);
        szSig = NULL;
        nRet = OPENSOAP_SEC_SIGNGEN_ERROR;
        goto FuncEnd;
    }
    /* Convert To Base64 String */
    nRet = openSOAPSecEncByBase64(szSig,
                                  (unsigned long)nSigLen,
                                  &szResBase64);
    if (OPENSOAP_FAILED(nRet)) {
        goto FuncEnd;
    }
    free(szSig);
    szSig = szResBase64;

FuncEnd:
    /* Terminate */
    if (pRsa != NULL)
        RSA_free(pRsa);
    if (szHash != NULL)
        free(szHash);
    *pszDSig = szSig;
    if (OPENSOAP_FAILED(nRet)) {
        if (*pszDSig != NULL) {
            free(*pszDSig);
            *pszDSig = NULL;
        }
        if (pszDigest != NULL) {
            if (*pszDigest != NULL) {
                free(*pszDigest);
                *pszDigest = NULL;
            }
        }
    }
    return nRet;
}
/***************************************************************************** 
    Function      : Generate Digital Signature(from Private Key File)(RSA)
    Return        : int
 ************************************************ Yuji Yamawaki 01.07.26 *****/
int
openSOAPSecMakeRSASignFromFile
(int                  iType,      /* (i)  Hash Type(OPENSOAP_HA_*) */
 const unsigned char* szData,     /* (i)  Data to be signed */
 unsigned long        ulDataLen,  /* (i)  Length of Data */
 const char*          szPrivKeyFileName, /* (i)  RSA Private Key File Name */
 char**               pszDigest,   /* (o)  Digest Value<base64> */
 char**               pszDSig)     /* (o)  Digital Signature<base64> */
{
    FILE*          fpPrivate = NULL;
    int            nRet = OPENSOAP_NO_ERROR;
    /* Check Arguments */
    if (szData == NULL || ulDataLen <= 0 || szPrivKeyFileName == NULL ||
        pszDSig == NULL) {
        return OPENSOAP_PARAMETER_BADVALUE;
    }
    /* Open Private Key File */
    if ((fpPrivate = fopen(szPrivKeyFileName, "r")) == NULL) {
        return OPENSOAP_FILEOPEN_ERROR;
    }
    /* Genetare Key File */
    nRet = openSOAPSecMakeRSASign(iType, szData, ulDataLen,
                                  fpPrivate, pszDigest, pszDSig);
    /* Terminate */
    fclose(fpPrivate);
    return nRet;
}

/***************************************************************************** 
    Function      : Verify Digital Signature(RSA)
    Return        : int
 ************************************************ Yuji Yamawaki 01.07.26 *****/
int
openSOAPSecVerifyRSASign
(int                  iType,          /* (i)  Hash Type(OPENSOAP_HA_*) */
 const char*          szIn,           /* (i)  Signature(Base64 string) */
 const unsigned char* szData,         /* (i)  Original Data */
 unsigned long        ulDataSize,     /* (i)  Original Data Size */
 FILE*                fpPubKey)       /* (i)  RSA Public Key File Stream */
{
    RSA*  pRsa = NULL;
    unsigned char* szHash = NULL;
    unsigned long  ulLenHash;
    unsigned char* szSig = NULL;  /* Signature(Binary) */
    unsigned long  ulSigLen;      /* Length of Signature(Binary) */
    int            nRet = OPENSOAP_NO_ERROR;
    int            iRes;
    fpos_t         fposKey;
    /* Check Arguments */
    if (szIn == NULL || szData == NULL || ulDataSize == 0 ||
        fpPubKey == NULL) {
        return OPENSOAP_PARAMETER_BADVALUE;
    }
    /* Backup Stream's Position */
    if (fgetpos(fpPubKey, &fposKey) != 0) {
        nRet = OPENSOAP_IO_READ_ERROR;
        goto FuncEnd;
    }
    /* Decode Base64 */
    nRet = openSOAPSecDecByBase64(szIn,
                                  &ulSigLen,
                                  &szSig);
    if (OPENSOAP_FAILED(nRet)) {
        goto FuncEnd;
    }
    /* Hash Original Data */
    nRet = openSOAPSecMakeHash(iType, szData, ulDataSize,
                               &ulLenHash, &szHash);
    if (OPENSOAP_FAILED(nRet)) {
        goto FuncEnd;
    }
    /* Generate RSA from Public key */
    if (PEM_read_RSAPublicKey(fpPubKey, &pRsa, NULL, NULL) == NULL) {
        nRet = OPENSOAP_SEC_SIGNVERIFY_ERROR;
        goto FuncEnd;
    }
    /* Recover File Position */
    if (fsetpos(fpPubKey, &fposKey) != 0) {
        nRet = OPENSOAP_IO_READ_ERROR;
        goto FuncEnd;
    }
    /* Verify Signature */
    iRes = RSA_verify(convType(iType),
                      szHash,
                      ulLenHash,
                      szSig,
                      (int)ulSigLen,
                      pRsa);
    if (iRes != 1) {
        nRet = OPENSOAP_SEC_SIGNVERIFY_ERROR;
    }
FuncEnd:
    if (szSig != NULL)
        free(szSig);
    if (szHash != NULL)
        free(szHash);
    /* Terminate */
    RSA_free(pRsa);
    return nRet;
}
/***************************************************************************** 
    Function      : Verify Digital Signature(Using Public Key File)(RSA)
    Return        : int
 ************************************************ Yuji Yamawaki 01.07.26 *****/
int
openSOAPSecVerifyRSASignFromFile
(int                  iType,            /* (i)  Hash Type(OPENSOAP_HA_*) */
 const char*          szIn,             /* (i)  Signature(Base64 string) */
 const unsigned char* szData,           /* (i)  Original Data */
 unsigned long        ulDataSize,       /* (i)  Original Data Size */
 const char*          szPubKeyFileName) /* (i)  RSA Public Key File Name */
{
    FILE* fpPublic = NULL;
    int   nRet = OPENSOAP_NO_ERROR;
    /* Check Arguments */
    if (szIn == NULL || szData == NULL || ulDataSize == 0 ||
        szPubKeyFileName == NULL) {
        return OPENSOAP_PARAMETER_BADVALUE;
    }
    /* Open Public Key File */
    if ((fpPublic = fopen(szPubKeyFileName, "r")) == NULL) {
        return OPENSOAP_SEC_SIGNVERIFY_ERROR;
    }
    /* Verify */
    nRet = openSOAPSecVerifyRSASign(iType, szIn, szData, ulDataSize,
                                   fpPublic);
    /* Terminate */
    fclose(fpPublic);
    return nRet;
}
