/* -*- mode: c; -*-
/*-----------------------------------------------------------------------------
 * $RCSfile: TransactionService.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <OpenSOAP/Service.h>
#include <OpenSOAP/Transport.h>

#include <stdio.h>

#if !defined(CONNECT_TYPE)
# define CONNECT_TYPE "stdio"
#endif /* CONNECT_TYPE */

static const char TRANSACTION_CONTROL_METHOD[] = "TransactionControl";
static const char TRANSACTION_HEADER_BLOCK[] = "TransactionHeaderBlock";
static const char TRANSACTION_BODY_BLOCK[] = "TransactionBodyBlock";
static const char TRANSACTION_ENDPOINT[] = "endpoint";
static const char TRANSACTION_REQUEST_ID_ATTRIBUTE[] = "transactionRequestID";

static const char TRANSACTION_NAMESPACE[] = "http://services.opensoap.jp/transaction/";
static const char TRANSACTION_PREFIX[] = "t";

static const char TRANSACTION_RESULT[] = "TransactionResult";
static const char TRANSACTION_ID_ATTRIBUTE[] = "transactionID";
static const char TRANSACTION_SUCCESS[] = "SUCCESS";
static const char TRANSACTION_ACTION_METHOD[] = "TransactionAction";
static const char TRANSACTION_ACTION_RESPONSE_BODY_BLOCK[] = "TransactionActionResponseBodyBlock";
static const char TRANSACTION_COMMIT[] = "COMMIT";
static const char TRANSACTION_ROLLBACK[] = "ROLLBACK";

static const char TRANSACTION_RESPONSE_METHOD[] = "TransactionResponse";

typedef struct{
    OpenSOAPStringPtr       endpoint;
    OpenSOAPXMLAttrPtr      trRequestIDAttr;

    OpenSOAPXMLElmPtr       requestHeaderBlock;
    OpenSOAPXMLElmPtr       requestBodyBlock;

    OpenSOAPEnvelopePtr     requestEnvelope;
    OpenSOAPEnvelopePtr     responseEnvelope;
    OpenSOAPXMLAttrPtr      trResponseIDAttr;

    OpenSOAPEnvelopePtr     actionRequestEnvelope;
    OpenSOAPEnvelopePtr     actionResponseEnvelope;
} TransactionParameters;


static
void
OutputEnvelope(OpenSOAPEnvelopePtr env,
	       const char *char_enc,
	       const char *msg,
	       FILE *fp) {

    if (!msg) {
        msg = "";
    }
    if (env && fp) {
        OpenSOAPByteArrayPtr env_buf = NULL;
	int error_code = OpenSOAPByteArrayCreate(&env_buf);
	if (OPENSOAP_SUCCEEDED(error_code)) {
	    error_code = OpenSOAPEnvelopeGetCharEncodingString(env,
							       char_enc,
							       env_buf);
	    if (OPENSOAP_SUCCEEDED(error_code)) {
	        const unsigned char *env_beg = NULL;
		size_t env_sz = 0;
		error_code = OpenSOAPByteArrayGetBeginSizeConst(env_buf,
								&env_beg,
								&env_sz);
		if (OPENSOAP_SUCCEEDED(error_code)) {
		    fprintf(fp,
			    "\n"
			    "=== %s soap envelope begin ===\n",
			    msg);
		    fwrite(env_beg, 1, env_sz, fp);
		    fprintf(fp,
			    "\n"
			    "=== %s soap envelope  end  ===\n",
			    msg);
		}
	    }
	    
	    OpenSOAPByteArrayRelease(env_buf);
	}
    }
}


void 
CheckStr(char *preStr, OpenSOAPStringPtr strPtr)
{
  char *string;
  size_t len = 0;
  OpenSOAPStringGetLengthMB(strPtr, &len);
  string = malloc(len + 1);
  OpenSOAPStringGetStringMB(strPtr, &len, string);
  fprintf(stderr,"%s: %s\n",preStr,string);
  free(string);
}


int 
CountProcessNo(OpenSOAPBlockPtr process)
{
    int trNo=0;
    OpenSOAPXMLElmPtr elm = NULL;
    OpenSOAPStringPtr elmName = NULL;
    int cmpRslt = 1;
  
    for(elm = NULL;
	OPENSOAP_SUCCEEDED(OpenSOAPBlockGetNextChild(process,&elm))
	&& elm;){
        OpenSOAPXMLElmGetNameString(elm, &elmName);
	OpenSOAPStringCompareMB(elmName,
				TRANSACTION_ENDPOINT,
				&cmpRslt);
	if(cmpRslt == 0){
	    trNo++;
	}
    }

    fprintf(stderr, "Number of Transactions = %d\n",trNo);
    return trNo;
}


static
int 
InitializeTransactionProcessParameter(TransactionParameters *trPrm,
				      int trNo)
{
  int i;
  for(i=0; i<trNo; i++){
    trPrm[i].endpoint               = NULL;
    trPrm[i].trRequestIDAttr        = NULL;

    trPrm[i].requestHeaderBlock     = NULL;
    trPrm[i].requestBodyBlock       = NULL;
    
    trPrm[i].requestEnvelope        = NULL;
    trPrm[i].responseEnvelope       = NULL;
    trPrm[i].trResponseIDAttr       = NULL;

    trPrm[i].actionRequestEnvelope  = NULL;
    trPrm[i].actionResponseEnvelope = NULL;
  }
  return 0;
}


static
int 
CheckTransactionParameter(TransactionParameters *trPrm,
			  int trNo)
{
  int i;
  OpenSOAPStringPtr attrStr = NULL;

  for(i=0; i<trNo; i++){
    fprintf(stderr, "Parameters[%d]\n",i);
    CheckStr("endpoint",trPrm[i].endpoint);
    OpenSOAPXMLAttrGetValueMB(trPrm[i].trRequestIDAttr,"string",&attrStr);
    CheckStr("trIDAttr",attrStr);
  }
 
}


static
int 
CriateTransactionProcessParameters(TransactionParameters *trPrm,
				   OpenSOAPEnvelopePtr request,
				   int trNo)
{
    int i;
    int errCode = 0;
    OpenSOAPBlockPtr bodyBlock = NULL;
    OpenSOAPXMLElmPtr elm      = NULL; 
    OpenSOAPStringPtr elmName  = NULL;
    int cmpRslt = 1;

    if(OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextBodyBlock(request,&bodyBlock))
       && bodyBlock){
    
        /* Transaction Proccess Section */ 
        for(i=0;
	    OPENSOAP_SUCCEEDED(OpenSOAPBlockGetNextChild(bodyBlock,&elm))
	    && elm
	    && i<trNo;){

	    /* Check Endpoint Element */
	    OpenSOAPXMLElmGetNameString(elm, &elmName);
	    OpenSOAPStringCompareMB(elmName, TRANSACTION_ENDPOINT, &cmpRslt);
	    if(cmpRslt == 0){

	        /* Getting each endpoint */
	        errCode = OpenSOAPXMLElmGetValueMB(
			      elm,
			      "string",
			      &trPrm[i].endpoint);

		/* Getting each request Transaction ID */
		errCode = OpenSOAPXMLElmGetAttributeMB(
			      elm,
			      TRANSACTION_REQUEST_ID_ATTRIBUTE,
			      &trPrm[i].trRequestIDAttr);
		i++;
	    }
	}
    }
    return errCode;
}


static
int
GetTransactionBlock(TransactionParameters *trPrm,
		    OpenSOAPEnvelopePtr request,
		    int trNo)
{
    int                  i = 0;
    int                  ret = 0;
    OpenSOAPBlockPtr     bodyBlock = NULL;
    OpenSOAPXMLAttrPtr   attrPtr = NULL;
    OpenSOAPStringPtr    attrStr = NULL;
    OpenSOAPStringPtr    attrPrm = NULL;
    OpenSOAPStringPtr    blockName = NULL;
    int                  attrCmpRslt = 1;
    int                  bodyCmpRslt = 1;
    int                  headerCmpRslt = 1;

    /* Transaction Body Block Section */
    for(i=0;i<trNo;i++){
        attrPtr = NULL;
	OpenSOAPXMLAttrGetValueMB(trPrm[i].trRequestIDAttr,"string",&attrPrm);
	for(bodyBlock=NULL;
	    OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextBodyBlock(request,
								&bodyBlock))
	    && bodyBlock;){

	    /* Attribute Name Check */
	    if(OPENSOAP_SUCCEEDED(
	         OpenSOAPBlockGetAttributeMB(bodyBlock,
					     TRANSACTION_REQUEST_ID_ATTRIBUTE,
					     &attrPtr))
	       && attrPtr!=NULL){
	
	        /* Attribute Value Check */
	        OpenSOAPXMLAttrGetValueMB(attrPtr,"string",&attrStr);
		OpenSOAPStringCompare(attrStr,attrPrm,&attrCmpRslt);
		if(attrCmpRslt == 0){

		    /* Bodyblock Name Check */
		    OpenSOAPBlockGetName(bodyBlock, &blockName);
		    OpenSOAPStringCompareMB(blockName,
					    TRANSACTION_HEADER_BLOCK,
					    &headerCmpRslt);
		    OpenSOAPStringCompareMB(blockName,
					    TRANSACTION_BODY_BLOCK,
					    &bodyCmpRslt);
		    if(headerCmpRslt == 0){
		        trPrm[i].requestHeaderBlock 
		            = (OpenSOAPXMLElmPtr)bodyBlock;
		    }
		    else if(bodyCmpRslt == 0){
		        trPrm[i].requestBodyBlock 
			    = (OpenSOAPXMLElmPtr)bodyBlock;
		    }
		    else ret = 1;
		}
	    }
	}
    }
    return ret;
}



static
int 
XMLElmDupricate(OpenSOAPXMLElmPtr elm,
		OpenSOAPXMLElmPtr dupElm)
{
    size_t                  len = 0;
    int                     i = 0;

    OpenSOAPXMLNamespacePtr ns = NULL;
    OpenSOAPStringPtr       nsURL = NULL;
    char *                  nsURLString = NULL;
    OpenSOAPStringPtr       nsPrefix = NULL;
    char *                  nsPrefixString = NULL;

    int                     attrNo = 0;
    OpenSOAPXMLAttrPtr      attr = NULL;
    OpenSOAPStringPtr       attrName = NULL;
    char *                  attrNameString = NULL;
    OpenSOAPStringPtr       attrValue = NULL;
    OpenSOAPXMLAttrPtr *    addedAttr = NULL;
    OpenSOAPXMLNamespacePtr attrNs = NULL;    
    OpenSOAPStringPtr       attrNsURL = NULL;
    char *                  attrNsURLString = NULL;
    OpenSOAPStringPtr       attrNsPrefix = NULL;
    char *                  attrNsPrefixString = NULL;

    OpenSOAPStringPtr       value = NULL;

    int                     elmNo = 0;
    OpenSOAPXMLElmPtr       child = NULL;
    OpenSOAPStringPtr       childName = NULL;
    char *                  childNameString = NULL;
    OpenSOAPXMLElmPtr *     elmChild = NULL;


    /* Add Namespaces */
    if(OPENSOAP_SUCCEEDED(OpenSOAPXMLElmGetNamespace(elm, &ns)) 
       && ns){

	OpenSOAPXMLNamespaceGetURI(ns,&nsURL);
	OpenSOAPStringGetLengthMB(nsURL, &len);
	nsURLString = malloc(len + 1);
	OpenSOAPStringGetStringMB(nsURL, &len, nsURLString);
	
	OpenSOAPXMLNamespaceGetPrefix(ns,&nsPrefix);
	OpenSOAPStringGetLengthMB(nsPrefix, &len);
	nsPrefixString = malloc(len + 1);
	OpenSOAPStringGetStringMB(nsPrefix, &len, nsPrefixString);
	
	OpenSOAPXMLElmSetNamespaceMB(dupElm,
				     nsURLString,
				     nsPrefixString);
	OpenSOAPStringClear(nsURL);
	OpenSOAPStringClear(nsPrefix);
	free(nsURLString);
	free(nsPrefixString);
    }
    

    /* Add Attributes */
    attrNo = 0;
    for(attr = NULL;
	OPENSOAP_SUCCEEDED(OpenSOAPXMLElmGetNextAttr(elm, &attr))
	&& attr;){
      attrNo++;
    }
    addedAttr = (OpenSOAPXMLAttrPtr*)malloc(attrNo *
					    sizeof(OpenSOAPXMLAttrPtr));

    for(attr = NULL;
	OPENSOAP_SUCCEEDED(OpenSOAPXMLElmGetNextAttr(elm, &attr))
	&& attr
	&& i<attrNo;
	i++){
        
        addedAttr[i] = NULL;

        OpenSOAPXMLAttrGetName(attr,&attrName);
	OpenSOAPStringGetLengthMB(attrName, &len);
	attrNameString = malloc(len + 1);
	OpenSOAPStringGetStringMB(attrName, &len, attrNameString);

	OpenSOAPXMLAttrGetValueMB(attr,"string",&attrValue);
	
	OpenSOAPXMLElmAddAttributeMB(dupElm,
				     attrNameString,
				     "string",
				     &attrValue,
				     &addedAttr[i]);

	/* Add Namespace of Attribute */
	if(OPENSOAP_SUCCEEDED(OpenSOAPXMLAttrGetNamespace(attr,
							  &attrNs)) 
	   && attrNs){

	    OpenSOAPXMLNamespaceGetURI(attrNs,&attrNsURL);
	    OpenSOAPStringGetLengthMB(attrNsURL, &len);
	    attrNsURLString = malloc(len + 1);
	    OpenSOAPStringGetStringMB(attrNsURL, &len, attrNsURLString);
	
	    OpenSOAPXMLNamespaceGetPrefix(attrNs,&attrNsPrefix);
	    OpenSOAPStringGetLengthMB(attrNsPrefix, &len);
	    attrNsPrefixString = malloc(len + 1);
	    OpenSOAPStringGetStringMB(attrNsPrefix, &len, attrNsPrefixString);
	
	    OpenSOAPXMLAttrSetNamespaceMB(addedAttr[i],
					 attrNsURLString,
					 attrNsPrefixString);
	    OpenSOAPStringClear(attrNsURL);
	    OpenSOAPStringClear(attrNsPrefix);
	    free(attrNsURLString);
	    free(attrNsPrefixString);
	}
	
	OpenSOAPStringClear(attrName);
	OpenSOAPStringClear(attrValue);
	free(attrNameString);
    }
    free(addedAttr);

    /* Add Value */
    if(OPENSOAP_SUCCEEDED(OpenSOAPXMLElmGetValueMB(elm,
						   "string",
						   &value))
       && value){
        OpenSOAPXMLElmSetValueMB(dupElm,"string",&value);
        OpenSOAPStringClear(value);
    }

    /* Add Children */
    elmNo = 0;
    for(child = NULL;
	OPENSOAP_SUCCEEDED(OpenSOAPXMLElmGetNextChild(elm,&child))
	&& child;){
      elmNo++;
    }
    elmChild = (OpenSOAPXMLElmPtr*)malloc(elmNo * sizeof(OpenSOAPXMLElmPtr));
    i = 0;
    for(child = NULL;
	OPENSOAP_SUCCEEDED(OpenSOAPXMLElmGetNextChild(elm, &child))
	&& child
	&& i<elmNo;
	i++){
        
        elmChild[i] = NULL;

        OpenSOAPXMLElmGetNameString(child,&childName);
	OpenSOAPStringGetLengthMB(childName, &len);
	childNameString = malloc(len + 1);
	OpenSOAPStringGetStringMB(childName, &len, childNameString);
	
	OpenSOAPXMLElmAddChildMB(dupElm,
				 childNameString,
				 &elmChild[i]);
	
	XMLElmDupricate(child,elmChild[i]);

    }
    free(elmChild);

}


static
int 
CriateEachRequestEnvelope(TransactionParameters *trPrm,
			  int trNo)
{
    int i;
    int errCode;
    size_t len = 0;
    OpenSOAPXMLElmPtr headerBlock = NULL;
    OpenSOAPXMLElmPtr bodyBlock = NULL;
    OpenSOAPStringPtr blockName  = NULL;
    char *blockNameString = NULL;
    OpenSOAPBlockPtr requestHeaderBlock = NULL;
    OpenSOAPBlockPtr requestBodyBlock = NULL;
    char envelopeName[50];

    for(i=0; i<trNo;i++){
        errCode = OpenSOAPEnvelopeCreateMB("1.1", 
					   NULL,
					   &trPrm[i].requestEnvelope);
	if(OPENSOAP_FAILED(errCode)){
	    return errCode;
	}

	for(headerBlock = NULL;
	    OPENSOAP_SUCCEEDED(OpenSOAPXMLElmGetNextChild(
				  trPrm[i].requestHeaderBlock,
				  &headerBlock))
	    && headerBlock
	    && trPrm[i].requestHeaderBlock;){

	    /* Create Header Block */
	    OpenSOAPXMLElmGetNameString(headerBlock,&blockName);

	    OpenSOAPStringGetLengthMB(blockName, &len);
	    blockNameString = malloc(len + 1);
	    OpenSOAPStringGetStringMB(blockName, &len, blockNameString);

	    OpenSOAPEnvelopeAddHeaderBlockMB(trPrm[i].requestEnvelope,
					     blockNameString,
					     &requestHeaderBlock);
	    XMLElmDupricate(headerBlock,
			    (OpenSOAPXMLElmPtr)requestHeaderBlock);
	}

	for(bodyBlock = NULL;
	    OPENSOAP_SUCCEEDED(OpenSOAPXMLElmGetNextChild(
				  trPrm[i].requestBodyBlock,
				  &bodyBlock))
	    && bodyBlock
	    && trPrm[i].requestBodyBlock;){

	    /* Create Body Block */
	    OpenSOAPXMLElmGetNameString(bodyBlock,&blockName);

	    OpenSOAPStringGetLengthMB(blockName, &len);
	    blockNameString = malloc(len + 1);
	    OpenSOAPStringGetStringMB(blockName, &len, blockNameString);

	    OpenSOAPEnvelopeAddBodyBlockMB(trPrm[i].requestEnvelope,
					   blockNameString,
					   &requestBodyBlock);
	    XMLElmDupricate(bodyBlock,
			    (OpenSOAPXMLElmPtr)requestBodyBlock);      
	}
	free(blockNameString);

	sprintf(envelopeName, "request[%d]",i);
	OutputEnvelope(trPrm[i].requestEnvelope,
		       "UTF-8",
		       envelopeName,
		       stderr);
    }
    return errCode;
}

static
int
SequentialInvokeEachRequestEnvelope(TransactionParameters *trPrm,
				    int trNo)
{
  
    int                    i = 0;
    size_t                 len = 0;
    OpenSOAPTransportPtr   transport[trNo];
    char                   envelopeName[50];
    char *                 endpointString;
    OpenSOAPBlockPtr       firstBodyBlock = NULL;
    OpenSOAPXMLElmPtr      transactionResultElm = NULL;
    OpenSOAPStringPtr      transactionResultStr = NULL;
    int                    successCmpRslt = 1;
    char                   transactionNo[20];

    while(i<trNo){
	OpenSOAPStringGetLengthMB(trPrm[i].endpoint, &len);
	endpointString = malloc(len + 1);
	OpenSOAPStringGetStringMB(trPrm[i].endpoint, &len, endpointString);

	fprintf(stderr, "\nendpoint[%d]:%s\n",i,endpointString);

        OpenSOAPTransportCreate(&transport[i]);
	OpenSOAPTransportSetURL(transport[i],
				endpointString);
	OpenSOAPTransportInvoke(transport[i],
				trPrm[i].requestEnvelope, 
				&trPrm[i].responseEnvelope);
	OpenSOAPTransportRelease(transport[i]);
	
	sprintf(envelopeName, "response[%d]",i);
	OutputEnvelope(trPrm[i].responseEnvelope,
		       "UTF-8",
		       envelopeName,
		       stderr);

	/*Judge SUCCESS */
	firstBodyBlock = NULL;
	if(OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextBodyBlock(
				  trPrm[i].responseEnvelope, 
				  &firstBodyBlock))
	   && firstBodyBlock){
	    if(OPENSOAP_SUCCEEDED(OpenSOAPBlockGetChildMB(
				  firstBodyBlock,
				  TRANSACTION_RESULT,
				  &transactionResultElm))
	       && transactionResultElm){
	        OpenSOAPXMLElmGetAttributeMB(transactionResultElm,
					     TRANSACTION_ID_ATTRIBUTE,
					     &trPrm[i].trResponseIDAttr);
		OpenSOAPXMLElmGetValueMB(transactionResultElm,
					 "string",
					 &transactionResultStr);
		sprintf(transactionNo, "transaction[%d]",i);
		CheckStr(transactionNo,transactionResultStr);
		OpenSOAPStringCompareMB(transactionResultStr,
					TRANSACTION_SUCCESS,
					&successCmpRslt);
		if(successCmpRslt == 0){
		  i++;
		}
		else{
		  fprintf(stderr, "transaction[%d] element failed\n",i);
		  return i+1;
		}
	    }
	    else{
	      fprintf(stderr, "transaction[%d] body failed\n",i);
	      return i+1;
	    }
	}
	else{
	  fprintf(stderr, "transaction[%d] envelope failed\n",i);
	  return i+1;
	}
    }
    return 0;
}


static
int
TransactionActionInvoke(TransactionParameters *trPrm,
			int trNo,
			int successCode)
{
  
    int                    i = 0;
    size_t                 len = 0;
    int                    errCode = 0;
    int                    transactionServices = 0;
    OpenSOAPStringPtr      transactionAction = NULL;
    OpenSOAPTransportPtr   transport[trNo];
    char                   envelopeName[50];
    char *                 endpointString;
    OpenSOAPBlockPtr       actionBlock = NULL;
    OpenSOAPStringPtr      transactionID = NULL;
    OpenSOAPXMLAttrPtr     trIDAttr = NULL;


    if(successCode == 0){
        transactionServices = trNo;
 	OpenSOAPStringCreateWithMB(TRANSACTION_COMMIT, &transactionAction);
    }
    else{
        transactionServices = successCode - 1;
 	OpenSOAPStringCreateWithMB(TRANSACTION_ROLLBACK, &transactionAction);
    }
    for(i=0; i<transactionServices; i++){

        /* Make Envelope of Transaction Action (COMMIT or ROLLBACK) */
        errCode = OpenSOAPEnvelopeCreateMB("1.1", 
					   NULL,
					   &trPrm[i].actionRequestEnvelope);
	if(OPENSOAP_FAILED(errCode)){
	    return errCode;
	}
	OpenSOAPEnvelopeAddBodyBlockMB(trPrm[i].actionRequestEnvelope,
				       TRANSACTION_ACTION_METHOD,
				       &actionBlock);

	/* Add Transaction ID Attribute */
	trIDAttr = NULL;
	errCode = OpenSOAPXMLAttrGetValueMB(trPrm[i].trResponseIDAttr,
					    "string",
					    &transactionID);
	if(OPENSOAP_SUCCEEDED(errCode)){
	    OpenSOAPBlockAddAttributeMB(actionBlock,
					TRANSACTION_ID_ATTRIBUTE,
					"string",
					&transactionID,
					&trIDAttr);
	}
	else return errCode;

	/* Add Namespace */
	OpenSOAPBlockSetNamespaceMB(actionBlock,
				    TRANSACTION_NAMESPACE,
				    TRANSACTION_PREFIX);

	/* Add Action Value */	
	OpenSOAPBlockSetValueMB(actionBlock,
				"string",
				&transactionAction);

	sprintf(envelopeName, "action request[%d]",i);
	OutputEnvelope(trPrm[i].actionRequestEnvelope,
		       "UTF-8",
		       envelopeName,
		       stderr);

	/* Invoke Envelope */
	OpenSOAPStringGetLengthMB(trPrm[i].endpoint, &len);
	endpointString = malloc(len + 1);
	OpenSOAPStringGetStringMB(trPrm[i].endpoint, &len, endpointString);

	fprintf(stderr, "\nendpoint[%d]:%s\n",i,endpointString);

        OpenSOAPTransportCreate(&transport[i]);
	OpenSOAPTransportSetURL(transport[i],
				endpointString);
	OpenSOAPTransportInvoke(transport[i],
				trPrm[i].actionRequestEnvelope, 
				&trPrm[i].actionResponseEnvelope);
	OpenSOAPTransportRelease(transport[i]);
	
	sprintf(envelopeName, "action response[%d]",i);
	OutputEnvelope(trPrm[i].actionResponseEnvelope,
		       "UTF-8",
		       envelopeName,
		       stderr);

    }
    return 0;
}


static
int
CreateResponseEnvelope(TransactionParameters *trPrm,
		       int trNo,
		       int successCode,
		       OpenSOAPEnvelopePtr *response)
{
    int                    i;
    int                    errCode;
    size_t                 len = 0;
    OpenSOAPStringPtr      transactionAction = NULL;
    OpenSOAPBlockPtr       responseBlock = NULL;
    OpenSOAPBlockPtr       responseHeaderBlock[trNo];
    OpenSOAPBlockPtr       responseBodyBlock[trNo];
    OpenSOAPBlockPtr       actionResponseBodyBlock[trNo];

    OpenSOAPStringPtr      trIDAttrValue  = NULL;
    OpenSOAPXMLAttrPtr     trIDAttr = NULL;

    OpenSOAPBlockPtr       headerBlocks = NULL;
    OpenSOAPBlockPtr       bodyBlocks = NULL;
    OpenSOAPBlockPtr       actionResponseBodyBlocks = NULL;
    OpenSOAPXMLElmPtr      addedHeaderBlocks = NULL;
    OpenSOAPXMLElmPtr      addedBodyBlocks = NULL;
    OpenSOAPXMLElmPtr      addedActionResponseBodyBlocks = NULL;

    OpenSOAPStringPtr      blockName  = NULL;
    char *                 blockNameString = NULL;


    if(successCode == 0){
 	OpenSOAPStringCreateWithMB(TRANSACTION_COMMIT, &transactionAction);
    }
    else{
 	OpenSOAPStringCreateWithMB(TRANSACTION_ROLLBACK, &transactionAction);
    }

    errCode = OpenSOAPEnvelopeCreateMB("1.1",  NULL, response);
    if(OPENSOAP_FAILED(errCode)){
      return errCode;
    }
    
    /* Add Final Result Response Block */
    OpenSOAPEnvelopeAddBodyBlockMB(*response,
				   TRANSACTION_RESPONSE_METHOD,
				   &responseBlock);
    OpenSOAPBlockSetNamespaceMB(responseBlock,
				TRANSACTION_NAMESPACE,
				TRANSACTION_PREFIX);
    OpenSOAPBlockSetValueMB(responseBlock,
			    "string",
			    &transactionAction);

    for(i=0; i<trNo; i++){
        responseHeaderBlock[i] = NULL;
        responseBodyBlock[i] = NULL;
	actionResponseBodyBlock[i] = NULL;
	OpenSOAPXMLAttrGetValueMB(trPrm[i].trRequestIDAttr,
				  "string",
				  &trIDAttrValue);

        /* Create Header Block */
	headerBlocks = NULL;
	if(OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextHeaderBlock(
				  trPrm[i].responseEnvelope,
				  &headerBlocks))
	   && headerBlocks
	   && trPrm[i].responseEnvelope){
	  
	    OpenSOAPEnvelopeAddBodyBlockMB(*response,
					   TRANSACTION_HEADER_BLOCK,
					   &responseHeaderBlock[i]);
	    OpenSOAPBlockSetNamespaceMB(responseHeaderBlock[i],
					TRANSACTION_NAMESPACE,
					TRANSACTION_PREFIX);
	    OpenSOAPBlockAddAttributeMB(responseHeaderBlock[i],
					TRANSACTION_REQUEST_ID_ATTRIBUTE,
					"string",
					&trIDAttrValue,
					&trIDAttr);
	    for(headerBlocks = NULL;
		OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextHeaderBlock(
				       trPrm[i].responseEnvelope,
				       &headerBlocks))
		&& headerBlocks;){

	        addedHeaderBlocks = NULL;
	        OpenSOAPBlockGetName(headerBlocks,&blockName);

		OpenSOAPStringGetLengthMB(blockName, &len);
		blockNameString = malloc(len + 1);
		OpenSOAPStringGetStringMB(blockName, &len, blockNameString);

		addedHeaderBlocks = NULL;
		OpenSOAPBlockAddChildMB(responseHeaderBlock[i],
					blockNameString,
					&addedHeaderBlocks);

		XMLElmDupricate((OpenSOAPXMLElmPtr)headerBlocks,
				addedHeaderBlocks);
	    }
	}

        /* Create Body Block */
	bodyBlocks = NULL;
	if(OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextBodyBlock(
				  trPrm[i].responseEnvelope,
				  &bodyBlocks))
	   && bodyBlocks
	   && trPrm[i].responseEnvelope){

	    OpenSOAPEnvelopeAddBodyBlockMB(*response,
					   TRANSACTION_BODY_BLOCK,
					   &responseBodyBlock[i]);
	    OpenSOAPBlockSetNamespaceMB(responseBodyBlock[i],
					TRANSACTION_NAMESPACE,
					TRANSACTION_PREFIX);
	    OpenSOAPBlockAddAttributeMB(responseBodyBlock[i],
					TRANSACTION_REQUEST_ID_ATTRIBUTE,
					"string",
					&trIDAttrValue,
					&trIDAttr);
	    for(bodyBlocks = NULL;
		OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextBodyBlock(
				       trPrm[i].responseEnvelope,
				       &bodyBlocks))
		&& bodyBlocks;){

	        addedBodyBlocks = NULL;
	        OpenSOAPBlockGetName(bodyBlocks,&blockName);

		OpenSOAPStringGetLengthMB(blockName, &len);
		blockNameString = malloc(len + 1);
		OpenSOAPStringGetStringMB(blockName, &len, blockNameString);

		addedBodyBlocks = NULL;
		OpenSOAPBlockAddChildMB(responseBodyBlock[i],
					blockNameString,
					&addedBodyBlocks);

		XMLElmDupricate((OpenSOAPXMLElmPtr)bodyBlocks,
				addedBodyBlocks);
	    }
	}

        /* Create Action Response Block */
	actionResponseBodyBlocks = NULL;
	if(OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextBodyBlock(
				  trPrm[i].actionResponseEnvelope,
				  &actionResponseBodyBlocks))
	   && actionResponseBodyBlocks
	   && trPrm[i].actionResponseEnvelope){

	    OpenSOAPEnvelopeAddBodyBlockMB(*response,
					   TRANSACTION_ACTION_RESPONSE_BODY_BLOCK,
					   &actionResponseBodyBlock[i]);
	    OpenSOAPBlockSetNamespaceMB(actionResponseBodyBlock[i],
					TRANSACTION_NAMESPACE,
					TRANSACTION_PREFIX);
	    OpenSOAPBlockAddAttributeMB(actionResponseBodyBlock[i],
					TRANSACTION_REQUEST_ID_ATTRIBUTE,
					"string",
					&trIDAttrValue,
					&trIDAttr);
	    for(actionResponseBodyBlocks = NULL;
		OPENSOAP_SUCCEEDED(OpenSOAPEnvelopeGetNextBodyBlock(
				       trPrm[i].actionResponseEnvelope,
				       &actionResponseBodyBlocks))
		&& actionResponseBodyBlocks;){

	        addedActionResponseBodyBlocks = NULL;
	        OpenSOAPBlockGetName(actionResponseBodyBlocks,&blockName);

		OpenSOAPStringGetLengthMB(blockName, &len);
		blockNameString = malloc(len + 1);
		OpenSOAPStringGetStringMB(blockName, &len, blockNameString);

		addedActionResponseBodyBlocks = NULL;
		OpenSOAPBlockAddChildMB(actionResponseBodyBlock[i],
					blockNameString,
					&addedActionResponseBodyBlocks);

		XMLElmDupricate((OpenSOAPXMLElmPtr)actionResponseBodyBlocks,
				addedActionResponseBodyBlocks);
	    }
	}
	free(blockNameString);

    }
    /* OutputEnvelope(*response,"UTF-8", "response", stderr); */

    return errCode;
}


static
int
TransactionServiceFunc(OpenSOAPEnvelopePtr request,
		       OpenSOAPEnvelopePtr *response,
		       void *opt) {
	
    OpenSOAPBlockPtr transactionProcess = NULL;

    TransactionParameters *trPrm;
    int trNo;

    int trSucceed;

    /* parse request message -process part-*/
	
    OpenSOAPEnvelopeGetBodyBlockMB(request,
				   TRANSACTION_CONTROL_METHOD,
				   &transactionProcess);
    trNo = CountProcessNo(transactionProcess);
    trPrm = (TransactionParameters*)malloc(trNo *
					   sizeof(TransactionParameters));
    InitializeTransactionProcessParameter(trPrm, trNo);
    CriateTransactionProcessParameters(trPrm, request, trNo);
    GetTransactionBlock(trPrm, request, trNo);

    CheckTransactionParameter(trPrm,trNo);

    CriateEachRequestEnvelope(trPrm, trNo);

    trSucceed = SequentialInvokeEachRequestEnvelope(trPrm, trNo);
    fprintf(stderr, "Transaction:: Success Code = %d\n", trSucceed);

    TransactionActionInvoke(trPrm, trNo, trSucceed);

    CreateResponseEnvelope(trPrm, trNo, trSucceed, response);

    free(trPrm);
	
    return 0;
}


int
main(void) {
	
    OpenSOAPServicePtr transaction_service = NULL;
	
    int error_code = OpenSOAPInitialize(NULL);
	
    if (OPENSOAP_SUCCEEDED(error_code)) {
        error_code = OpenSOAPServiceCreateMB(&transaction_service,
					     "TransactionService",
					     CONNECT_TYPE,
					     0);
	
	if (OPENSOAP_SUCCEEDED(error_code)) {
	    error_code = OpenSOAPServiceRegisterMB(transaction_service,
						   TRANSACTION_CONTROL_METHOD,
						   TransactionServiceFunc,
						   NULL);
	}
	if (OPENSOAP_SUCCEEDED(error_code)) {
	    error_code = OpenSOAPServiceRun(transaction_service);
	}
	
	OpenSOAPServiceRelease(transaction_service);
	
	OpenSOAPUltimate();
    }

    return 0;
}
